% Generated by roxygen2 (4.1.0): do not edit by hand
% Please edit documentation in R/tigger_wrapper.R
\name{runTigger}
\alias{runTigger}
\title{Infer genotype (including novel alleles) and correct V calls}
\usage{
runTigger(sample_db, germline_db, find_novel = TRUE, find_genotype = TRUE,
  correct_calls = TRUE, allele_min = 1e-04, y_intercept = 1/8,
  nt_min = 1, nt_max = 312, mut_min = 1, mut_max = 10, j_max = 0.1,
  min_seqs = 50, min_frac = 3/4, fraction_to_explain = 7/8,
  gene_cutoff = 0.001, seq_gap = "SEQUENCE_IMGT", v_call_col = "V_CALL",
  j_call_col = "J_CALL", junc_length_col = "JUNCTION_LENGTH",
  quiet = FALSE)
}
\arguments{
\item{sample_db}{a data frame with the colums described in
Details below.}

\item{germline_db}{a vector of named nucleotide germline
sequences matching the calls in
\code{sample_db}}

\item{find_novel}{logical. Should novel alleles be searched for?}

\item{find_genotype}{logical. Should the genotype be inferred?}

\item{correct_calls}{logical. Should the allele calls be corrected?}

\item{allele_min}{a number < 1 representing the minimum fraction
of sequences required for an allele to not be
excluded from analysis. or a number >= 1
representing the minimum count for sequences.
See \code{\link{assignAlleleGroups}}.}

\item{y_intercept}{the y-intercept above which positions should
be considered potentially polymorphic. See
\code{\link{detectNovelV}}.}

\item{nt_min}{the first nucleotide position to be considered
in intercept calculations. See
\code{\link{detectNovelV}}.}

\item{nt_max}{the last nucleotide position to be considered
in intercept calculations. See
\code{\link{detectNovelV}}.}

\item{mut_min}{the minimum number of mutations carried by
sequences used in in intercept calculations.
See \code{\link{detectNovelV}}.}

\item{mut_max}{the maximum number of mutations carried by
sequences used in in intercept calculations.
See \code{\link{detectNovelV}}.}

\item{j_max}{the maximum fraction of sequences perfectly
aligning to a potential novel allele that are
allowed to utilize to a particular combination
of junction length and J gene. See
\code{\link{detectNovelV}}.}

\item{min_seqs}{the minimum number of total sequences (within
the desired mutational range and nucleotide
range) required for the samples to be
analyzed for polymorphisms. See
\code{\link{detectNovelV}}.}

\item{min_frac}{the maxmium number of total sequences (within
the desired mutational range and nucleotide
range) required for the samples to be
analyzed for polymorphisms. See
\code{\link{detectNovelV}}.}

\item{fraction_to_explain}{the portion of each gene that must be
explained by the alleles that will be included
in the genotype. See
\code{\link{inferGenotype}}.}

\item{gene_cutoff}{the minimum fraction of the unmutated
sequences that must be attributed to a gene
in order for it to be included in the
genotype. See \code{\link{inferGenotype}}.}

\item{seq_gap}{the name of the column in \code{sample_db}
that includes the IMGT-gapped sequence}

\item{v_call_col}{the name of the column in \code{sample_db}
that includes the intial V call
in the column indicated by \code{seq_gap}}

\item{j_call_col}{the name of the column in \code{sample_db}
that includes the initial J call}

\item{junc_length_col}{the name of the column in \code{sample_db}
that includes the junction length}

\item{quiet}{logical indicating if additional diagonostic
                               output will be suppressed}

\item{v_length_col}{the name of the column in \code{sample_db}
that includes the length of the V sequence
contained within \code{seq_gap}}
}
\value{
a list containing data on new alleles, the inferred genotype, and
the corrected IGHV calls.
}
\description{
\code{runTigger} takes a table of sample sequences from a single subject and a
vector of database germline sequences. It then performs the following:
(1) Infers the presence of novel IGHV alleles not in the germline database.
(2) Infers the individuals V genotype.
(3) Corrects the IGHV allele calls of the samples based on the IGHV genotype.
The sample sequences should be a data frame where each row is a sequence and
each column contains data about that sequence. The database germlines should
be a vector of sequences with names matching those in the table of sample
sequences.
}
\details{
The required columns that must be contained within \code{sample_db}
are detailed below:
\itemize{
\item{\code{SEQUENCE_IMGT}: }{V(D)J sequence in the IMGT gapped format}
\item{\code{V_CALL}: }{(Comma separated) name(s) of the nearest V allele(s)}
\item{\code{J_CALL}: }{(Comma separated) name(s) of the nearest J allele(s)}
\item{\code{JUNCTION_LENGTH}: }{Length of the junction region of the V(D)J sample}
}
}
\examples{
\dontrun{
## Load example data and run all aspects of TIgGER (takes a few minutes)
data(sample_db)
data(germline_ighv)
results = runTigger(sample_db, germline_ighv)

## Summarize the detected novel alleles, add them to vector of all alleles
novel_sequences = novelSummary(results, seqs_to_return = "in genotype")
germline_ighv = c(germline_ighv, novel_sequences)
## Plot positional mutation frequency versus sequence-wide mutation count
plotNovelLines(results$novel)
## Plot nucleotide usage at polymorphic positions
plotNovelBars(results$novel)
## Plot J and junction usage for sequences perfectly matching novel alleles
plotJunctionBars(results$novel)

## View the inferred genotype
print(results$genotype)
## Get the nucleotide sequences of all genotype alleles
genotype_sequences = genotypeFasta(results$genotype, germline_ighv)

## Extract the corrected V allele calls and appened them to the data frame
V_CALL_GENOTYPED = results$new_calls
sample_db = cbind(sample_db, V_CALL_GENOTYPED)
}
}
\references{
Gadala-Maria D, Yaari G, Uduman M, Kleinstein SH (2015) Automated
analysis of high-throughput B cell sequencing data reveals a high frequency
of novel immunoglobulin V gene segment alleles. PNAS. 112(8):E862-70
}
\seealso{
\code{\link{detectNovelV}}, \code{\link{inferGenotype}},
         \code{\link{reassignAlleles}}
}

